@@ -23,8 +23,17 @@ limitations under the License.
 
 static char *jsonnet_str(struct JsonnetVm *vm, const char *str)
 {
-    char *out = jsonnet_realloc(vm, NULL, strlen(str) + 1);
-    memcpy(out, str, strlen(str) + 1);
+    size_t size = strlen(str) + 1;
+    char *out = jsonnet_realloc(vm, NULL, size);
+    memcpy(out, str, size);
+    return out;
+}
+
+static char *jsonnet_str_nonull(struct JsonnetVm *vm, const char *str, size_t *buflen)
+{
+    *buflen = strlen(str);
+    char *out = jsonnet_realloc(vm, NULL, *buflen);
+    memcpy(out, str, *buflen);
     return out;
 }
 
@@ -42,6 +51,7 @@ static const char *exc_to_str(void)
 
 struct NativeCtx {
     struct JsonnetVm *vm;
+    PyThreadState **py_thread;
     PyObject *callback;
     size_t argc;
 };
@@ -139,6 +149,8 @@ static struct JsonnetJsonValue *cpython_native_callback(
     const struct NativeCtx *ctx = ctx_;
     int i;
 
+    PyEval_RestoreThread(*ctx->py_thread);
+
     PyObject *arglist;  // Will hold a tuple of strings.
     PyObject *result;  // Will hold a string.
 
@@ -168,6 +180,7 @@ static struct JsonnetJsonValue *cpython_native_callback(
             // TODO(dcunnin): Support objects (to dicts).
             Py_DECREF(arglist);
             *succ = 0;
+            *ctx->py_thread = PyEval_SaveThread();
             return jsonnet_json_make_string(ctx->vm, "Non-primitive param.");
         }
         PyTuple_SetItem(arglist, i, pyobj);
@@ -182,6 +195,7 @@ static struct JsonnetJsonValue *cpython_native_callback(
         struct JsonnetJsonValue *r = jsonnet_json_make_string(ctx->vm, exc_to_str());
         *succ = 0;
         PyErr_Clear();
+        *ctx->py_thread = PyEval_SaveThread();
         return r;
     }
 
@@ -193,67 +207,75 @@ static struct JsonnetJsonValue *cpython_native_callback(
         *succ = 0;
         r = jsonnet_json_make_string(ctx->vm, err_msg);
     }
+    *ctx->py_thread = PyEval_SaveThread();
     return r;
 }
 
 
 struct ImportCtx {
     struct JsonnetVm *vm;
+    PyThreadState **py_thread;
     PyObject *callback;
 };
 
-static char *cpython_import_callback(void *ctx_, const char *base, const char *rel,
-                                     char **found_here, int *success)
+static int cpython_import_callback(void *ctx_, const char *base, const char *rel,
+                                   char **found_here, char **buf, size_t *buflen)
 {
     const struct ImportCtx *ctx = ctx_;
     PyObject *arglist, *result;
-    char *out;
+    int success;
 
+    PyEval_RestoreThread(*ctx->py_thread);
     arglist = Py_BuildValue("(s, s)", base, rel);
     result = PyEval_CallObject(ctx->callback, arglist);
     Py_DECREF(arglist);
 
     if (result == NULL) {
         // Get string from exception
-        char *out = jsonnet_str(ctx->vm, exc_to_str());
-        *success = 0;
+        *buf = jsonnet_str_nonull(ctx->vm, exc_to_str(), buflen);
         PyErr_Clear();
-        return out;
+        *ctx->py_thread = PyEval_SaveThread();
+        return 1; // failure
     }
 
     if (!PyTuple_Check(result)) {
-        out = jsonnet_str(ctx->vm, "import_callback did not return a tuple");
-        *success = 0;
+        *buf = jsonnet_str_nonull(ctx->vm, "import_callback did not return a tuple", buflen);
+        success = 0;
     } else if (PyTuple_Size(result) != 2) {
-        out = jsonnet_str(ctx->vm, "import_callback did not return a tuple (size 2)");
-        *success = 0;
+        *buf = jsonnet_str_nonull(ctx->vm, "import_callback did not return a tuple (size 2)", buflen);
+        success = 0;
     } else {
         PyObject *file_name = PyTuple_GetItem(result, 0);
         PyObject *file_content = PyTuple_GetItem(result, 1);
 #if PY_MAJOR_VERSION >= 3
-        if (!PyUnicode_Check(file_name) || !PyUnicode_Check(file_content)) {
+        if (!PyUnicode_Check(file_name) || !PyBytes_Check(file_content)) {
 #else
         if (!PyString_Check(file_name) || !PyString_Check(file_content)) {
 #endif
-            out = jsonnet_str(ctx->vm, "import_callback did not return a pair of strings");
-            *success = 0;
+            *buf = jsonnet_str_nonull(ctx->vm, "import_callback did not return (string, bytes)", buflen);
+            success = 0;
         } else {
+            const char *content_buf;
+            const ssize_t content_len;
 #if PY_MAJOR_VERSION >= 3
             const char *found_here_cstr = PyUnicode_AsUTF8(file_name);
-            const char *content_cstr = PyUnicode_AsUTF8(file_content);
+            PyBytes_AsStringAndSize(file_content, &content_buf, &content_len);
 #else
             const char *found_here_cstr = PyString_AsString(file_name);
-            const char *content_cstr = PyString_AsString(file_content);
+            PyString_AsStringAndSize(file_content, &content_buf, &content_len);
 #endif
             *found_here = jsonnet_str(ctx->vm, found_here_cstr);
-            out = jsonnet_str(ctx->vm, content_cstr);
-            *success = 1;
+            *buflen = content_len - 1; // Python always adds a trailing null
+            *buf = jsonnet_realloc(ctx->vm, NULL, *buflen);
+            memcpy(*buf, content_buf, *buflen);
+            success = 1;
         }
     }
 
     Py_DECREF(result);
+    *ctx->py_thread = PyEval_SaveThread();
 
-    return out;
+    return success ? 0 : 1;
 }
 
 static PyObject *handle_result(struct JsonnetVm *vm, char *out, int error)
@@ -340,7 +362,7 @@ int handle_import_callback(struct ImportCtx *ctx, PyObject *import_callback)
  * \returns 1 on success, 0 with exception set upon failure.
  */
 static int handle_native_callbacks(struct JsonnetVm *vm, PyObject *native_callbacks,
-                                   struct NativeCtx **ctxs)
+                                   struct NativeCtx **ctxs, PyThreadState **py_thread)
 {
     size_t num_natives = 0;
     PyObject *key, *val;
@@ -433,6 +455,7 @@ static int handle_native_callbacks(struct JsonnetVm *vm, PyObject *native_callba
         }
         params_c[num_params] = NULL;
         (*ctxs)[num_natives].vm = vm;
+        (*ctxs)[num_natives].py_thread = py_thread;
         (*ctxs)[num_natives].callback = PyTuple_GetItem(val, 1);
         (*ctxs)[num_natives].argc = num_params;
         jsonnet_native_callback(vm, key_, cpython_native_callback, &(*ctxs)[num_natives],
@@ -448,7 +471,8 @@ static int handle_native_callbacks(struct JsonnetVm *vm, PyObject *native_callba
 static PyObject* evaluate_file(PyObject* self, PyObject* args, PyObject *keywds)
 {
     const char *filename;
-    char *out, *jpath_str;
+    char *out;
+    const char *jpath_str;
     unsigned max_stack = 500, gc_min_objects = 1000, max_trace = 20;
     double gc_growth_trigger = 2;
     int error;
@@ -478,6 +502,8 @@ static PyObject* evaluate_file(PyObject* self, PyObject* args, PyObject *keywds)
         return NULL;
     }
 
+    PyThreadState *py_thread;
+
     vm = jsonnet_make();
     jsonnet_max_stack(vm, max_stack);
     jsonnet_gc_min_objects(vm, gc_min_objects);
@@ -515,16 +541,19 @@ static PyObject* evaluate_file(PyObject* self, PyObject* args, PyObject *keywds)
     if (!handle_vars(vm, ext_codes, 1, 0)) return NULL;
     if (!handle_vars(vm, tla_vars, 0, 1)) return NULL;
     if (!handle_vars(vm, tla_codes, 1, 1)) return NULL;
-    struct ImportCtx ctx = { vm, import_callback };
+
+    struct ImportCtx ctx = { vm, &py_thread, import_callback };
     if (!handle_import_callback(&ctx, import_callback)) {
         return NULL;
     }
     struct NativeCtx *ctxs = NULL;
-    if (!handle_native_callbacks(vm, native_callbacks, &ctxs)) {
+    if (!handle_native_callbacks(vm, native_callbacks, &ctxs, &py_thread)) {
         free(ctxs);
         return NULL;
     }
+    py_thread = PyEval_SaveThread();
     out = jsonnet_evaluate_file(vm, filename, &error);
+    PyEval_RestoreThread(py_thread);
     free(ctxs);
     return handle_result(vm, out, error);
 }
@@ -532,7 +561,8 @@ static PyObject* evaluate_file(PyObject* self, PyObject* args, PyObject *keywds)
 static PyObject* evaluate_snippet(PyObject* self, PyObject* args, PyObject *keywds)
 {
     const char *filename, *src;
-    char *out, *jpath_str;
+    char *out;
+    const char *jpath_str;
     unsigned max_stack = 500, gc_min_objects = 1000, max_trace = 20;
     double gc_growth_trigger = 2;
     int error;
@@ -562,6 +592,8 @@ static PyObject* evaluate_snippet(PyObject* self, PyObject* args, PyObject *keyw
         return NULL;
     }
 
+    PyThreadState *py_thread;
+
     vm = jsonnet_make();
     jsonnet_max_stack(vm, max_stack);
     jsonnet_gc_min_objects(vm, gc_min_objects);
@@ -599,16 +631,18 @@ static PyObject* evaluate_snippet(PyObject* self, PyObject* args, PyObject *keyw
     if (!handle_vars(vm, ext_codes, 1, 0)) return NULL;
     if (!handle_vars(vm, tla_vars, 0, 1)) return NULL;
     if (!handle_vars(vm, tla_codes, 1, 1)) return NULL;
-    struct ImportCtx ctx = { vm, import_callback };
+    struct ImportCtx ctx = { vm, &py_thread, import_callback };
     if (!handle_import_callback(&ctx, import_callback)) {
         return NULL;
     }
     struct NativeCtx *ctxs = NULL;
-    if (!handle_native_callbacks(vm, native_callbacks, &ctxs)) {
+    if (!handle_native_callbacks(vm, native_callbacks, &ctxs, &py_thread)) {
         free(ctxs);
         return NULL;
     }
+    py_thread = PyEval_SaveThread();
     out = jsonnet_evaluate_snippet(vm, filename, src, &error);
+    PyEval_RestoreThread(py_thread);
     free(ctxs);
     return handle_result(vm, out, error);
 }
